from unittest.mock import patch

import pytest

from constellaxion.handlers.cloud_job import AWSDeployJob, GCPDeployJob
from constellaxion.handlers.dataset import Dataset
from constellaxion.handlers.model import Model
from constellaxion.handlers.training import Training


@pytest.fixture
def valid_model():
    """Test model with valid configuration."""
    return Model("test-model", "tiny-llama-1b")


@pytest.fixture
def valid_dataset():
    """Test dataset with valid configuration."""
    return Dataset("train.csv", "val.csv", "test.csv", "test-model")


@pytest.fixture
def valid_training():
    """Test training with valid configuration."""
    return Training(3, 32)


def test_gcp_deploy_job_create_config_with_invalid_model():
    """Test GCP deploy job configuration creation with invalid model."""
    with pytest.raises(AttributeError):
        GCPDeployJob.create_config(
            None,
            "test-project",
            "us-central1",
            "test@test-project.iam.gserviceaccount.com",
            None,
            None,
        )


def test_aws_deploy_job_create_config_with_invalid_model():
    """Test AWS deploy job configuration creation with invalid model."""
    with pytest.raises(AttributeError):
        AWSDeployJob.create_config(None, "us-east-1", None, None)


def test_gcp_deploy_job_create_config_with_missing_project_id(
    valid_model, valid_dataset, valid_training
):
    """Test GCP deploy job configuration creation with missing project ID."""
    with pytest.raises(ValueError):
        GCPDeployJob.create_config(
            valid_model,
            "",
            "us-central1",
            "test@test-project.iam.gserviceaccount.com",
            valid_dataset,
            valid_training,
        )


def test_aws_deploy_job_create_config_with_missing_region(
    valid_model, valid_dataset, valid_training
):
    """Test AWS deploy job configuration creation with missing region."""
    with pytest.raises(ValueError):
        AWSDeployJob.create_config(valid_model, "", valid_dataset, valid_training)


@patch("constellaxion.services.gcp.gcp_deploy_job.run_gcp_deploy_job")
def test_gcp_deploy_job_deploy_with_invalid_config(_mock_run_gcp_deploy_job):
    """Test GCP deploy job deployment with invalid config."""
    with pytest.raises(KeyError):
        GCPDeployJob.deploy({})


@patch("constellaxion.services.aws.aws_deploy_job.run_aws_deploy_job")
def test_aws_deploy_job_deploy_with_invalid_config(_mock_run_aws_deploy_job):
    """Test AWS deploy job deployment with invalid config."""
    with pytest.raises(KeyError):
        AWSDeployJob.deploy({})


@patch("constellaxion.services.gcp.prompt_gcp_model.send_gcp_prompt")
def test_gcp_deploy_job_prompt_with_invalid_config(_mock_send_gcp_prompt):
    """Test GCP deploy job prompt with invalid config."""
    with pytest.raises(KeyError):
        GCPDeployJob.prompt("Test prompt", {})


@patch("constellaxion.services.aws.prompt_aws_model.send_aws_prompt")
def test_aws_deploy_job_prompt_with_invalid_config(_mock_send_aws_prompt):
    """Test AWS deploy job prompt with invalid config."""
    with pytest.raises(KeyError):
        AWSDeployJob.prompt("Test prompt", {})



import json
import os

import click

from constellaxion.handlers.cloud_job import AWSDeployJob, GCPDeployJob


def get_job(show=False):
    """Load and optionally print the job configuration from job.json."""
    if os.path.exists("job.json"):
        with open("job.json", "r", encoding="utf-8") as f:
            config = json.load(f)
        if show:
            click.echo(click.style("Model Job Config Details:", bold=True, fg="blue"))
            click.echo(json.dumps(config, indent=4))
        return config
    else:
        click.echo(
            click.style(
                "Error: job.json not found. Run 'constellaxion init' first", fg="red"
            )
        )
        return None


@click.group()
def model():
    """Manage model jobs"""
    pass


@model.command()
def prompt():
    """Prompt a deployed model"""
    config = get_job()
    cloud = config["deploy"]["provider"]
    model_id = config["model"]["model_id"]
    click.clear()  # Clear the screen
    click.echo(click.style(f"Send a prompt to {model_id}", fg="yellow", bold=True))
    click.echo(click.style("Type 'exit' or 'quit' to quit"))
    while True:
        if cloud and config["deploy"]["endpoint_path"]:
            response = ""
            click.echo(click.style("\nPrompt: ", fg="green"), nl=False)
            txt = input()
            if txt.lower() in ["exit", "quit"]:
                break

            if cloud == "gcp":
                job = GCPDeployJob()
                response = job.prompt(txt, config)
            elif cloud == "aws":
                job = AWSDeployJob()
                response = job.prompt(txt, config)
            click.echo(click.style(f"\n🤖 {model_id}: ", fg="green") + response)
        else:
            click.echo(
                click.style(
                    "Error: Trained model not found. Try training and deploying a model first",
                    fg="red",
                )
            )
            break


@model.command()
def train():
    """Run training job"""
    click.echo(click.style("Preparing training job...", fg="blue"))
    config = get_job()
    if config:
        cloud = config["deploy"]["provider"]
        if cloud == "gcp":
            job = GCPDeployJob()
            job.run(config)
        # elif cloud == "aws":
        #     job = AWSDeployJob()
        #     job.run(config)


@model.command(help="Serve a trained model")
def serve():
    """Serve Model"""
    config = get_job()
    if config:
        model_id = config["model"]["model_id"]
        click.echo(click.style(f"Serving model with ID: {model_id}", fg="blue"))
        cloud = config["deploy"]["provider"]
        if cloud == "gcp":
            job = GCPDeployJob()
            job.serve(config)
        elif cloud == "aws":
            job = AWSDeployJob()
            job.serve(config)


@model.command()
def deploy():
    """Deploy a model"""
    click.echo(click.style("Deploying model...", fg="blue"))
    job_config = get_job()
    cloud = job_config["deploy"]["provider"]
    if cloud == "gcp":
        job = GCPDeployJob()
        job.deploy(job_config)
    elif cloud == "aws":
        job = AWSDeployJob()
        job.deploy(job_config)


@model.command()
def view():
    """View the status or details of one or more jobs"""
    get_job(show=True)



import pytest

from constellaxion.handlers.model import Model


def test_model_initialization():
    """Test that a Model can be initialized with valid parameters."""
    model = Model("test-model", "tiny-llama-1b")
    assert model.id == "test-model"  # nosec: B101
    assert model.base_model == "tiny-llama-1b"  # nosec: B101


def test_model_empty_base_model():
    """Test that Model raises an error with invalid base model."""
    with pytest.raises(ValueError):
        Model("test-model", "")


def test_model_empty_id():
    """Test that Model raises an error with empty ID."""
    with pytest.raises(ValueError):
        Model("", "tiny-llama-1b")


def test_model_none_id():
    """Test that Model raises an error with None ID."""
    with pytest.raises(ValueError):
        Model(None, "tiny-llama-1b")


def test_model_none_base_model():
    """Test that Model raises an error with None base model."""
    with pytest.raises(ValueError):
        Model("test-model", None)



class Set:
    """Set class for handling dataset paths."""

    def __init__(self, dataset_type: str, path: str):
        if not path:
            raise ValueError(f"Path for {dataset_type} set cannot be empty")
        self.type = dataset_type
        self.path = path


class Dataset:
    """Dataset class for handling dataset paths."""

    def __init__(self, train: str, val: str, test: str, model_id: str):
        if not train or not val or not test:
            raise ValueError("All dataset paths must be provided")
        if not model_id:
            raise ValueError("Model ID must be provided")
        self._train = Set("train", train)
        self._val = Set("val", val)
        self._test = Set("test", test)
        self.model_id = model_id
        self.train = self._get_train()
        self.val = self._get_val()
        self.test = self._get_test()

    def _get_train(self):
        """Get the train path."""
        return self._train.path

    def _get_val(self):
        """Get the val path."""
        return self._val.path

    def _get_test(self):
        """Get the test path."""
        return self._test.path

    def to_dict(self):
        """Convert the dataset to a dictionary."""
        return {
            "train": {"local": self.train, "cloud": f"{self.model_id}/data/train.csv"},
            "val": {"local": self.val, "cloud": f"{self.model_id}/data/val.csv"},
            "test": {"local": self.test, "cloud": f"{self.model_id}/data/test.csv"},
        }



import pytest

from constellaxion.handlers.dataset import Dataset, Set


def test_set_initialization():
    """Test that a Set can be initialized with valid parameters."""
    dataset_set = Set("train", "train.csv")
    assert dataset_set.type == "train"  # nosec: B101
    assert dataset_set.path == "train.csv"  # nosec: B101


def test_dataset_initialization():
    """Test that a Dataset can be initialized with valid parameters."""
    dataset = Dataset("train.csv", "val.csv", "test.csv", "test-model")
    assert dataset.train == "train.csv"  # nosec: B101
    assert dataset.val == "val.csv"  # nosec: B101
    assert dataset.test == "test.csv"  # nosec: B101
    assert dataset.model_id == "test-model"  # nosec: B101


def test_dataset_to_dict():
    """Test that to_dict returns the correct dictionary structure."""
    dataset = Dataset("train.csv", "val.csv", "test.csv", "test-model")
    expected = {
        "train": {"local": "train.csv", "cloud": "test-model/data/train.csv"},
        "val": {"local": "val.csv", "cloud": "test-model/data/val.csv"},
        "test": {"local": "test.csv", "cloud": "test-model/data/test.csv"},
    }
    assert dataset.to_dict() == expected  # nosec: B101


def test_dataset_empty_paths():
    """Test that Dataset raises an error with empty paths."""
    with pytest.raises(ValueError):
        Dataset("", "val.csv", "test.csv", "test-model")
    with pytest.raises(ValueError):
        Dataset("train.csv", "", "test.csv", "test-model")
    with pytest.raises(ValueError):
        Dataset("train.csv", "val.csv", "", "test-model")


def test_dataset_none_paths():
    """Test that Dataset raises an error with None paths."""
    with pytest.raises(ValueError):
        Dataset(None, "val.csv", "test.csv", "test-model")
    with pytest.raises(ValueError):
        Dataset("train.csv", None, "test.csv", "test-model")
    with pytest.raises(ValueError):
        Dataset("train.csv", "val.csv", None, "test-model")


def test_dataset_empty_model_id():
    """Test that Dataset raises an error with empty model_id."""
    with pytest.raises(ValueError):
        Dataset("train.csv", "val.csv", "test.csv", "")


def test_dataset_none_model_id():
    """Test that Dataset raises an error with None model_id."""
    with pytest.raises(ValueError):
        Dataset("train.csv", "val.csv", "test.csv", None)



from click.testing import CliRunner
import pytest

from constellaxion.commands.init import init, init_dataset, init_model, init_training
from constellaxion.handlers.dataset import Dataset
from constellaxion.handlers.model import Model
from constellaxion.handlers.training import Training


@pytest.fixture
def valid_model_config():
    """Test model initialization with valid config."""
    return {
        "id": "test-model",
        "base": "tiny-llama-1b",
    }


@pytest.fixture
def valid_dataset_config():
    """Test dataset initialization with valid config."""
    return {
        "train": "./train.csv",
        "val": "./val.csv",
        "test": "./test.csv",
    }


@pytest.fixture
def valid_training_config():
    """Test training initialization with valid config."""
    return {
        "epochs": 3,
        "batch_size": 32,
    }


def test_init_model_valid_config(valid_model_config):
    """Test model initialization with valid config."""
    model = init_model(valid_model_config)
    print(model.id)
    print(model.base_model)
    assert isinstance(model, Model)  # nosec: B101
    assert model.id == "test-model"  # nosec: B101
    assert model.base_model == "tiny-llama-1b"  # nosec: B101


def test_init_model_invalid_config():
    """Test model initialization with invalid config."""
    with pytest.raises(AttributeError):
        init_model({})


def test_init_dataset_valid_config(valid_dataset_config, valid_model_config):
    """Test dataset initialization with valid config."""
    dataset = init_dataset(valid_dataset_config, valid_model_config)
    assert isinstance(dataset, Dataset)  # nosec: B101
    assert dataset.train == "./train.csv"  # nosec: B101
    assert dataset.val == "./val.csv"  # nosec: B101
    assert dataset.test == "./test.csv"  # nosec: B101
    assert dataset.model_id == "test-model"  # nosec: B101


def test_init_dataset_invalid_config(valid_model_config):
    """Test dataset initialization with invalid config."""
    with pytest.raises(AttributeError):
        init_dataset({}, valid_model_config)


def test_init_training_valid_config(valid_training_config):
    """Test training initialization with valid config."""
    training = init_training(valid_training_config)
    assert isinstance(training, Training)  # nosec: B101
    assert training.epochs == 3  # nosec: B101
    assert training.batch_size == 32  # nosec: B101


def test_init_training_invalid_config():
    """Test training initialization with invalid config."""
    with pytest.raises(AttributeError):
        init_training({})


def test_init_command_with_invalid_yaml():
    """Test init command with invalid YAML file."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        # Create an invalid YAML file
        with open("model.yaml", "w", encoding="utf-8") as f:
            f.write("invalid: yaml: content: [")

        result = runner.invoke(init)
        assert result.exit_code == 1  # nosec: B101
        assert "mapping values are not allowed here" in result.output  # nosec: B101


def test_init_command_with_missing_yaml():
    """Test init command with missing YAML file."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(init)
        assert result.exit_code == 1  # nosec: B101
        assert "Error: model.yaml file not found" in result.output  # nosec: B101



"""Utility functions for the constellaxion CLI."""

import contextlib
import json
import logging


def get_json(path):
    """Get job configuration"""
    with open(path, "r", encoding="utf-8") as f:
        j = json.load(f)
        return j


@contextlib.contextmanager
def suppress_logs_and_warnings():
    """Context manager to temporarily suppress all logs and warnings."""
    logging.basicConfig(level=logging.CRITICAL)
    logging.getLogger("sagemaker.config").setLevel(logging.CRITICAL)
    logging.getLogger("boto3").setLevel(logging.CRITICAL)
    logging.getLogger("botocore").setLevel(logging.CRITICAL)
    logging.getLogger("pydantic").setLevel(logging.CRITICAL)


# Function to get the effective level name
def get_level_name(level):
    """Get the effective level name for a logging level."""
    if level == logging.DEBUG:
        return "DEBUG"
    elif level == logging.INFO:
        return "INFO"
    elif level == logging.WARNING:
        return "WARNING"
    elif level == logging.ERROR:
        return "ERROR"
    elif level == logging.CRITICAL:
        return "CRITICAL"
    else:
        return "NOTSET"


def check_logging_levels():
    """Check the logging levels of all loggers."""
    # Get the root logger
    root_logger = logging.getLogger()
    # Get all other loggers
    loggers = logging.Logger.manager.loggerDict
    with open("log.txt", "w", encoding="utf-8") as f:
        f.write(f"Logger: root, Level: {get_level_name(root_logger.level)}\n")
        for name, logger in loggers.items():
            if isinstance(logger, logging.PlaceHolder):
                continue
            f.write(f"Logger: {name}, Level: {get_level_name(logger.level)}\n")



"""
This module provides functions for managing IAM roles and service accounts in GCP.
"""

from google.auth import default
from google.auth.exceptions import RefreshError
from google.auth.transport.requests import Request
from google.oauth2.id_token import verify_oauth2_token
from googleapiclient.discovery import Resource, build


def create_service_account(project_id: str) -> str:
    """
    Create a service account and assign roles.

    Args:
        project_id (str): GCP Project ID.
    """
    roles = ["roles/aiplatform.user", "roles/storage.admin"]
    service_account_email = f"constellaxion-admin@{project_id}.iam.gserviceaccount.com"

    try:
        # Get credentials
        credentials, _ = default(
            scopes=["https://www.googleapis.com/auth/cloud-platform"]
        )

        iam_service: Resource = build("iam", "v1", credentials=credentials)

        # Extract the email from credentials info
        user_email = get_logged_in_user_email()

        # Create service account
        try:
            iam_service.projects().serviceAccounts().create(
                name=f"projects/{project_id}",
                body={
                    "accountId": "constellaxion-admin",
                    "serviceAccount": {
                        "displayName": "Constellaxion Admin",
                    },
                },
            ).execute()
            print(f"Service account created: {service_account_email}")
        except Exception as e:
            # Handle specific errors
            if hasattr(e, "resp") and e.resp.status == 409:
                print(
                    "ConstellaXion Admin Service Account already exists. Continuing..."
                )
            else:
                raise

        # Assign project-level roles
        assign_project_roles(project_id, service_account_email, roles)

        # Allow user to act as the service account
        assign_impersonation_role(project_id, service_account_email, user_email)

        return service_account_email

    except RefreshError as e:
        print(
            "Error: Could not refresh credentials. Please ensure you are authenticated."
        )
        print(f"Details: {e}")
        raise
    except Exception as e:
        print(f"Unexpected error: {e}")
        raise


def get_logged_in_user_email():
    """
    Get the email of the currently authenticated user.

    Returns:
        str: The email of the authenticated user.
    """
    try:
        # Get default credentials
        credentials, _ = default()

        # Refresh the credentials to get a valid token
        credentials.refresh(Request())

        # Get the ID token
        if not hasattr(credentials, "id_token") or not credentials.id_token:
            raise ValueError("No ID token found in credentials.")

        # Decode the ID token to extract the email
        id_info = verify_oauth2_token(credentials.id_token, Request())
        email = id_info.get("email")
        if not email:
            raise ValueError("Email not found in ID token.")

        return email
    except Exception as e:
        print(f"Error retrieving user email: {e}")
        raise


def assign_project_roles(project_id, service_account_email, roles):
    """
    Assign project-level roles to a service account, skipping roles that are already assigned.

    Args:
        project_id (str): GCP Project ID.
        service_account_email (str): Email of the service account.
        roles (list): List of roles to assign.
    """
    credentials, _ = default(scopes=["https://www.googleapis.com/auth/cloud-platform"])
    cloud_resource_manager = build(
        "cloudresourcemanager", "v1", credentials=credentials
    )

    # Get the current IAM policy for the project
    policy = (
        cloud_resource_manager.projects()
        .getIamPolicy(resource=project_id, body={})
        .execute()
    )

    # Extract existing bindings
    bindings = policy.get("bindings", [])
    for role in roles:
        # Check if the role is already assigned
        binding = next((b for b in bindings if b["role"] == role), None)
        if binding:
            members = binding.get("members", [])
            if f"serviceAccount:{service_account_email}" in members:
                print(
                    f"Role {role} already assigned to {service_account_email}. Skipping..."
                )
                continue
            else:
                # Add the service account to the existing binding
                binding["members"].append(f"serviceAccount:{service_account_email}")
        else:
            # Create a new binding if the role is not in the policy
            bindings.append(
                {"role": role, "members": [f"serviceAccount:{service_account_email}"]}
            )

    # Update the policy with the modified bindings
    policy["bindings"] = bindings

    # Push the updated policy back to GCP
    cloud_resource_manager.projects().setIamPolicy(
        resource=project_id, body={"policy": policy}
    ).execute()

    print(f"Roles assigned successfully: {roles}")


def assign_impersonation_role(project_id, service_account_email, user_email):
    """
    Assign the Service Account User role to a user to allow them to act as the service account.

    Args:
        project_id (str): GCP Project ID.
        service_account_email (str): Email of the service account.
        user_email (str): Email of the user to allow acting as the service account.
    """
    credentials, _ = default(scopes=["https://www.googleapis.com/auth/cloud-platform"])
    iam_service = build("iam", "v1", credentials=credentials)

    # Get the current IAM policy for the service account
    resource = f"projects/{project_id}/serviceAccounts/{service_account_email}"
    policy = (
        iam_service.projects()
        .serviceAccounts()
        .getIamPolicy(resource=resource)
        .execute()
    )

    # Check if the user already has the role
    bindings = policy.get("bindings", [])
    role = "roles/iam.serviceAccountUser"
    binding = next((b for b in bindings if b["role"] == role), None)
    if binding:
        members = binding.get("members", [])
        if f"user:{user_email}" in members:
            print(
                f"User {user_email} already has 'roles/iam.serviceAccountUser' on {service_account_email}. Skipping..."
            )
            return
        else:
            binding["members"].append(f"user:{user_email}")
    else:
        # Add a new binding for the role
        bindings.append({"role": role, "members": [f"user:{user_email}"]})

    # Update the policy with the modified bindings
    policy["bindings"] = bindings
    iam_service.projects().serviceAccounts().setIamPolicy(
        resource=resource, body={"policy": policy}
    ).execute()

    print(
        f"User {user_email} granted 'roles/iam.serviceAccountUser' on {service_account_email}."
    )



import json
import logging

import pytest

from constellaxion.utils import get_json, get_level_name


def test_get_level_name():
    """Test that get_level_name returns the correct level names."""
    assert get_level_name(logging.DEBUG) == "DEBUG"  # nosec: B101
    assert get_level_name(logging.INFO) == "INFO"  # nosec: B101
    assert get_level_name(logging.WARNING) == "WARNING"  # nosec: B101
    assert get_level_name(logging.ERROR) == "ERROR"  # nosec: B101
    assert get_level_name(logging.CRITICAL) == "CRITICAL"  # nosec: B101
    assert get_level_name(999) == "NOTSET"  # nosec: B101


def test_get_json(tmp_path):
    """Test that get_json correctly reads JSON files."""
    # Create a temporary JSON file
    test_json = {"test": "data"}
    json_path = tmp_path / "test.json"
    with open(json_path, "w", encoding="utf-8") as f:
        json.dump(test_json, f)
    # Test reading the JSON file
    result = get_json(json_path)
    assert result == test_json  # nosec: B101


def test_get_json_invalid_file():
    """Test that get_json raises an error for invalid files."""
    with pytest.raises(FileNotFoundError):
        get_json("invalid_file.json")



from google.cloud import aiplatform


def send_gcp_prompt(prompt: str, endpoint_path: str, location: str):
    """Sends a request to a Vertex AI endpoint and returns the response."""
    # The AI Platform services require regional API endpoints.
    client_options = {"api_endpoint": f"{location}-aiplatform.googleapis.com"}
    # Initialize Vertex AI client
    client = aiplatform.gapic.PredictionServiceClient(client_options=client_options)

    instances = [{"prompt": prompt}]

    # Make prediction request
    request = {
        "endpoint": endpoint_path,
        "instances": instances,
    }

    response = client.predict(**request)

    return response.predictions[0]["prediction"]



"""Module for handling cloud deployment jobs across different providers (AWS, GCP)."""

from abc import ABC, abstractmethod
import json

from constellaxion.handlers.dataset import Dataset
from constellaxion.handlers.model import Model
from constellaxion.handlers.training import Training
from constellaxion.services.aws.aws_deploy_job import run_aws_deploy_job
from constellaxion.services.aws.prompt_aws_model import send_aws_prompt
from constellaxion.services.gcp.gcp_deploy_job import run_gcp_deploy_job
from constellaxion.services.gcp.prompt_gcp_model import send_gcp_prompt
from constellaxion.services.gcp.serve_job import run_serving_job
from constellaxion.services.gcp.train_job import run_training_job


class BaseCloudJob(ABC):
    """Base class for cloud deployment jobs providing common interface for model deployment."""

    def __init__(self):
        pass

    @staticmethod
    @abstractmethod
    def run(config):
        """Run model finetuning on GCP"""
        pass

    @staticmethod
    @abstractmethod
    def create_config(config):
        """Create a JSON configuration file from model and dataset attributes."""
        pass


class GCPDeployJob(BaseCloudJob):
    """GCP deployment job class."""

    @staticmethod
    def run(config):
        """Run model finetuning on GCP"""
        run_training_job(config)

    @staticmethod
    def serve(config):
        """Serve finetuned model on GCP"""
        endpoint_path = run_serving_job(config)
        config["deploy"]["endpoint_path"] = endpoint_path
        with open("job.json", "w", encoding="utf-8") as f:
            json.dump(config, f, indent=4)

    @staticmethod
    def deploy(config):
        """Deploy foundation model to GCP"""
        endpoint_path = run_gcp_deploy_job(config)
        config["deploy"]["endpoint_path"] = endpoint_path
        with open("job.json", "w", encoding="utf-8") as f:
            json.dump(config, f, indent=4)

    @staticmethod
    def prompt(prompt, config):
        """Send prompt to model"""
        endpoint_path = config["deploy"]["endpoint_path"]
        region = config["deploy"]["region"]
        response = send_gcp_prompt(prompt, endpoint_path, region)
        return response

    @staticmethod
    def create_config(
        model: Model,
        project_id: str,
        region: str,
        service_account: str,
        dataset: Dataset,
        training: Training,
    ):
        """Create a JSON configuration file from model and dataset attributes."""
        if not project_id or not region or not service_account:
            raise ValueError(
                "model, project_id, region, and service_account must be provided"
            )
        bucket_name = f"constellaxion-{project_id}"
        job_config = {
            "model": {
                "model_id": model.id,
                "base_model": model.base_model,
            },
            "dataset": dataset.to_dict() if dataset else None,
            "training": training.to_dict() if training else None,
            "deploy": {
                "provider": "gcp",
                "project_id": project_id,
                "region": region,
                "bucket_name": bucket_name,
                "staging_dir": f"{model.id}/staging",
                "experiments_dir": f"{model.id}/experiments",
                "model_path": f"{model.id}/model",
                "service_account": service_account,
            },
        }
        with open("job.json", "w", encoding="utf-8") as f:
            json.dump(job_config, f, indent=4)


class AWSDeployJob(BaseCloudJob):
    """AWS deployment job class."""

    @staticmethod
    def run(*args, **kwargs):
        """Run model finetuning on AWS"""

    @staticmethod
    def serve(*args, **kwargs):
        """Serve finetuned model on AWS"""

    @staticmethod
    def deploy(config):
        """Deploy foundation model to AWS"""
        endpoint_path = run_aws_deploy_job(config)
        config["deploy"]["endpoint_path"] = endpoint_path
        with open("job.json", "w", encoding="utf-8") as f:
            json.dump(config, f, indent=4)

    @staticmethod
    def prompt(prompt, config):
        """Send prompt to model"""
        endpoint_path = config["deploy"]["endpoint_path"]
        region = config["deploy"]["region"]
        response = send_aws_prompt(prompt, endpoint_path, region)
        return response

    @staticmethod
    def create_config(model: Model, region: str, dataset: Dataset, training: Training):
        """Create a JSON configuration file from model and dataset attributes."""
        if not region:
            raise ValueError("region must be provided")
        job_config = {
            "model": {
                "model_id": model.id,
                "base_model": model.base_model,
            },
            "dataset": dataset.to_dict() if dataset else None,
            "training": training.to_dict() if training else None,
            "deploy": {
                "provider": "aws",
                "region": region,
                "bucket_name": f"constellaxion-{model.id}",
                "staging_dir": f"{model.id}/staging",
                "experiments_dir": f"{model.id}/experiments",
                "model_path": f"{model.id}/model",
                "iam_role": "constellaxion-admin",
            },
        }
        with open("job.json", "w", encoding="utf-8") as f:
            json.dump(job_config, f, indent=4)



import json
from unittest.mock import patch

from click.testing import CliRunner
import pytest

from constellaxion.commands.model import get_job, model


@pytest.fixture
def valid_job_config():
    """Test job configuration."""
    return {
        "model": {"model_id": "test-model", "base_model": "tiny-llama-1b"},
        "deploy": {
            "provider": "gcp",
            "endpoint_path": "projects/test/locations/us-central1/endpoints/test-endpoint",
            "region": "us-central1",
        },
    }


@pytest.fixture
def runner():
    """Click test runner."""
    return CliRunner()


def test_get_job_with_missing_file(runner):
    """Test get_job when job.json doesn't exist."""
    with runner.isolated_filesystem():
        result = get_job()
        assert result is None  # nosec: B101


def test_get_job_with_valid_file(runner, valid_job_config):
    """Test get_job with valid job.json."""
    with runner.isolated_filesystem():
        with open("job.json", "w", encoding="utf-8") as f:
            json.dump(valid_job_config, f)
        result = get_job()
        assert result == valid_job_config  # nosec: B101


# @patch("constellaxion.handlers.cloud_job.GCPDeployJob.prompt")
# def test_model_prompt_command_gcp(mock_prompt, runner, valid_job_config):
#     """Test model prompt command with GCP provider."""
#     mock_prompt.return_value = "Test response"
#     with runner.isolated_filesystem():
#         with open("job.json", "w", encoding="utf-8") as f:
#             json.dump(valid_job_config, f)
#         with patch("builtins.input", return_value="test prompt"):
#             result = runner.invoke(model, ["prompt"])
#             assert result.exit_code == 0
#             mock_prompt.assert_called_once_with("test prompt", valid_job_config)


# @patch("constellaxion.handlers.cloud_job.AWSDeployJob.prompt")
# def test_model_prompt_command_aws(mock_prompt, runner, valid_job_config):
#     """Test model prompt command with AWS provider."""
#     mock_prompt.return_value = "Test response"
#     valid_job_config["deploy"]["provider"] = "aws"
#     with runner.isolated_filesystem():
#         with open("job.json", "w") as f:
#             json.dump(valid_job_config, f)
#         with patch("builtins.input", return_value="test prompt"):
#             result = runner.invoke(model, ["prompt"])
#             assert result.exit_code == 0
#             mock_prompt.assert_called_once_with("test prompt", valid_job_config)


# def test_model_prompt_command_without_endpoint(runner, valid_job_config):
#     """Test model prompt command when endpoint is not set."""
#     valid_job_config["deploy"]["endpoint_path"] = ""
#     with runner.isolated_filesystem():
#         with open("job.json", "w") as f:
#             json.dump(valid_job_config, f)
#         result = runner.invoke(model, ["prompt"])
#         assert result.exit_code == 0
#         assert "Error: Trained model not found" in result.output


@patch("constellaxion.handlers.cloud_job.GCPDeployJob.run")
def test_model_train_command_gcp(mock_run, runner, valid_job_config):
    """Test model train command with GCP provider."""
    with runner.isolated_filesystem():
        with open("job.json", "w", encoding="utf-8") as f:
            json.dump(valid_job_config, f)
        result = runner.invoke(model, ["train"])
        assert result.exit_code == 0  # nosec: B101
        mock_run.assert_called_once_with(valid_job_config)


@patch("constellaxion.handlers.cloud_job.GCPDeployJob.serve")
def test_model_serve_command_gcp(mock_serve, runner, valid_job_config):
    """Test model serve command with GCP provider."""
    with runner.isolated_filesystem():
        with open("job.json", "w", encoding="utf-8") as f:
            json.dump(valid_job_config, f)
        result = runner.invoke(model, ["serve"])
        assert result.exit_code == 0  # nosec: B101
        mock_serve.assert_called_once_with(valid_job_config)


@patch("constellaxion.handlers.cloud_job.AWSDeployJob.serve")
def test_model_serve_command_aws(mock_serve, runner, valid_job_config):
    """Test model serve command with AWS provider."""
    valid_job_config["deploy"]["provider"] = "aws"
    with runner.isolated_filesystem():
        with open("job.json", "w", encoding="utf-8") as f:
            json.dump(valid_job_config, f)
        result = runner.invoke(model, ["serve"])
        assert result.exit_code == 0  # nosec: B101
        mock_serve.assert_called_once_with(valid_job_config)


@patch("constellaxion.handlers.cloud_job.GCPDeployJob.deploy")
def test_model_deploy_command_gcp(mock_deploy, runner, valid_job_config):
    """Test model deploy command with GCP provider."""
    with runner.isolated_filesystem():
        with open("job.json", "w", encoding="utf-8") as f:
            json.dump(valid_job_config, f)
        result = runner.invoke(model, ["deploy"])
        assert result.exit_code == 0  # nosec: B101
        mock_deploy.assert_called_once_with(valid_job_config)


@patch("constellaxion.handlers.cloud_job.AWSDeployJob.deploy")
def test_model_deploy_command_aws(mock_deploy, runner, valid_job_config):
    """Test model deploy command with AWS provider."""
    valid_job_config["deploy"]["provider"] = "aws"
    with runner.isolated_filesystem():
        with open("job.json", "w", encoding="utf-8") as f:
            json.dump(valid_job_config, f)
        result = runner.invoke(model, ["deploy"])
        assert result.exit_code == 0  # nosec: B101
        mock_deploy.assert_called_once_with(valid_job_config)



"""AWS LMI deployment module for deploying models to SageMaker endpoints using Large Model Inference."""

import boto3
import sagemaker
from sagemaker.djl_inference.model import DJLModel

from constellaxion.models.model_map import model_map
from constellaxion.services.aws.utils import get_aws_account_id


def create_model_from_lmi_container(
    base_model: str, env_vars: dict, execution_role: str
):
    """Creates a SageMaker model using the LMI container."""
    model = DJLModel(model_id=base_model, env=env_vars, role=execution_role)
    return model


def deploy_model_to_endpoint(model, model_id: str, instance_type: str):
    """Deploys a model to a SageMaker endpoint using LMI."""
    endpoint_name = sagemaker.utils.name_from_base(model_id)
    predictor = model.deploy(
        initial_instance_count=1,
        instance_type=instance_type,
        endpoint_name=endpoint_name,
    )
    return predictor


def run_aws_deploy_job(config):
    """Runs the LMI deployment job by creating and deploying a model to SageMaker."""
    base_model_alias = config["model"]["base_model"]
    model_id = config["model"]["model_id"]
    region = config["deploy"]["region"]
    iam_role = config["deploy"]["iam_role"]
    account_id = get_aws_account_id()
    role_arn = f"arn:aws:iam::{account_id}:role/{iam_role}"
    infra_config = model_map[base_model_alias]["aws_infra"]
    base_model = model_map[base_model_alias]["base_model"]
    # Use the LMI container image
    # image_uri = (
    #     "763104351884.dkr.ecr.us-west-2.amazonaws.com/"
    #     "djl-inference:0.25.0-lmi-deepspeed0.10.0-cu118"
    # )
    instance_type = infra_config["instance_type"]
    accelerator_count = infra_config.get("accelerator_count", 1)
    dtype = "float16" if not infra_config.get("dtype") else infra_config.get("dtype")

    # LMI specific environment variables
    env_vars = {
        "MODEL_ID": base_model,
        "DTYPE": dtype,
        "OPTION_MODEL_LOADING_TIMEOUT": "3600",
        "OPTION_ROLLING_BATCH": "lmi-dist",
        "OPTION_MAX_ROLLING_BATCH_SIZE": "32",
        "OPTION_TENSOR_PARALLEL_DEGREE": str(accelerator_count),
        "OPTION_LOAD_IN_8BIT": "true" if dtype == "int8" else "false",
        "OPTION_LOAD_IN_4BIT": "true" if dtype == "int4" else "false",
    }

    boto3.setup_default_session(region_name=region)

    # Register the model with LMI container
    model = create_model_from_lmi_container(base_model, env_vars, role_arn)

    # Deploy to endpoint
    predictor = deploy_model_to_endpoint(model, model_id, instance_type)
    endpoint_name = predictor.endpoint_name
    return endpoint_name



"""Configuration mapping for supported foundation models and their infrastructure requirements."""

model_map = {
    "@tinyllama-1b": {
        "base_model": "TinyLlama/TinyLlama-1.1B-intermediate-step-1431k-3T",
        "dtype": None,
        "gcp_infra": {
            "machine_type": "g2-standard-4",
            "accelerator_type": "NVIDIA_L4",
            "accelerator_count": 1,
            "replica_count": 1,
            "images": {
                "serve": "europe-west2-docker.pkg.dev/constellaxion/serving-images/foundation-model:latest",
                "finetune": "",
            },
        },
        "aws_infra": {},
    },
    "@tinyllama-bnb-4bit": {
        "base_model": "unsloth/tinyllama-bnb-4bit",
        "dtype": None,
        "max_seq_length": 4096,
        "gcp_infra": {
            "machine_type": "g2-standard-8",
            "accelerator_type": "NVIDIA_L4",
            "accelerator_count": 1,
            "replica_count": 1,
            "images": {
                "serve": "europe-west2-docker.pkg.dev/constellaxion/serving-images/foundation-model:latest",
                "finetune": "europe-west2-docker.pkg.dev/constellaxion/finetuning/unsloth-base-model:latest",
            },
        },
        "aws_infra": {
            "instance_type": "ml.g4dn.xlarge",
            "accelerator_type": "NVIDIA_TESLA_T4",
            "accelerator_count": 1,
            "min_replica_count": 1,
            "max_replica_count": 2,
            "images": {
                "serve": "567326753429.dkr.ecr.eu-north-1.amazonaws.com/constellaxion/serving-images:foundation-model-v1",
                "finetune": "",
            },
        },
    },
    "@Mistral-7B-v0.1": {
        "base_model": "mistralai/Mistral-7B-v0.1",
        "gcp_infra": {
            "machine_type": "n1-highmem-8",
            "accelerator_type": "NVIDIA_TESLA_T4",
            "accelerator_count": 1,
            "replica_count": 1,
            "dtype": "float16",
        },
        "images": {
            "serve": "europe-west2-docker.pkg.dev/constellaxion/serving-images/foundation-model:latest",
            "finetune": "",
        },
        "aws_infra": {
            "instance_type": "ml.g4dn.xlarge",
            "accelerator_type": "NVIDIA_TESLA_T4",
            "accelerator_count": 1,
            "min_replica_count": 1,
            "max_replica_count": 2,
            "dtype": "float16",
            "images": {
                "serve": "public.ecr.aws/n6b1s4s0/constellaxion/serving:foundation-model-v1",
                "finetune": "",
            },
        },
    },
    "@Mixtral-8x7B": {
        "base_model": "mistralai/Mixtral-8x7B",
        "gcp_infra": {
            "machine_type": "n1-highmem-16",
            "accelerator_type": "NVIDIA_A100_40GB",
            "accelerator_count": 2,
            "replica_count": 1,
        },
        "images": {
            "serve": "europe-west2-docker.pkg.dev/constellaxion/serving-images/foundation-model",
            "finetune": "",
        },
        "aws_infra": {
            "instance_type": "ml.p4d.24xlarge",
            "accelerator_type": "NVIDIA_A100_40GB",
            "accelerator_count": 2,
            "min_replica_count": 1,
            "max_replica_count": 2,
            "images": {
                "serve": "public.ecr.aws/n6b1s4s0/constellaxion/serving:foundation-model-v1",
                "finetune": "",
            },
        },
    },
    "@Qwen1.5-7B": {
        "base_model": "Qwen/Qwen1.5-7B",
        "gcp_infra": {
            "machine_type": "n1-highmem-8",
            "accelerator_type": "NVIDIA_TESLA_T4",
            "accelerator_count": 1,
            "replica_count": 1,
            "images": {
                "serve": "europe-west2-docker.pkg.dev/constellaxion/serving-images/foundation-model:latest",
                "finetune": "",
            },
        },
        "aws_infra": {
            "instance_type": "ml.g4dn.xlarge",
            "accelerator_type": "NVIDIA_TESLA_T4",
            "accelerator_count": 1,
            "min_replica_count": 1,
            "max_replica_count": 2,
            "images": {
                "serve": "public.ecr.aws/n6b1s4s0/constellaxion/serving:foundation-model-v1",
                "finetune": "",
            },
        },
    },
    "@Qwen1.5-14B": {
        "base_model": "Qwen/Qwen1.5-14B",
        "gcp_infra": {
            "machine_type": "n1-highmem-16",
            "accelerator_type": "NVIDIA_A100_40GB",
            "accelerator_count": 2,
            "replica_count": 1,
            "images": {
                "serve": "europe-west2-docker.pkg.dev/constellaxion/serving-images/foundation-model:latest",
                "finetune": "",
            },
        },
        "aws_infra": {
            "instance_type": "ml.p4d.24xlarge",
            "accelerator_type": "NVIDIA_A100_40GB",
            "accelerator_count": 2,
            "min_replica_count": 1,
            "max_replica_count": 2,
            "images": {
                "serve": "public.ecr.aws/n6b1s4s0/constellaxion/serving:foundation-model-v1",
                "finetune": "",
            },
        },
    },
    "@Llama-2-7b-chat-hf": {
        "base_model": "meta-llama/Llama-2-7b-chat-hf",
        "gcp_infra": {
            "machine_type": "g2-standard-8",
            "accelerator_type": "NVIDIA_L4",
            "accelerator_count": 1,
            "replica_count": 1,
            "images": {
                "serve": "europe-west2-docker.pkg.dev/constellaxion/serving-images/foundation-model:latest",
                "finetune": "",
            },
        },
        "aws_infra": {
            "instance_type": "ml.g5.xlarge",
            "accelerator_type": "NVIDIA_L4",
            "accelerator_count": 1,
            "min_replica_count": 1,
            "max_replica_count": 2,
            "images": {
                "serve": "public.ecr.aws/n6b1s4s0/constellaxion/serving:foundation-model-v1",
                "finetune": "",
            },
        },
    },
    "@Llama-2-13b-hf": {
        "base_model": "meta-llama/Llama-2-13b-hf",
        "gcp_infra": {
            "machine_type": "n1-highmem-16",
            "accelerator_type": "NVIDIA_A100_40GB",
            "accelerator_count": 1,
            "replica_count": 1,
            "images": {
                "serve": "europe-west2-docker.pkg.dev/constellaxion/serving-images/foundation-model:latest",
                "finetune": "",
            },
        },
        "aws_infra": {
            "instance_type": "ml.p4d.24xlarge",
            "accelerator_type": "NVIDIA_A100_40GB",
            "accelerator_count": 1,
            "min_replica_count": 1,
            "max_replica_count": 2,
            "images": {
                "serve": "public.ecr.aws/n6b1s4s0/constellaxion/serving:foundation-model-v1",
                "finetune": "",
            },
        },
    },
    "@Llama-2-70b-chat-hf": {
        "base_model": "meta-llama/Llama-2-70b-chat-hf",
        "gcp_infra": {
            "machine_type": "n1-highmem-64",
            "accelerator_type": "NVIDIA_A100_80GB",
            "accelerator_count": 8,
            "replica_count": 1,
        },
        "images": {
            "serve": "europe-west2-docker.pkg.dev/constellaxion/serving-images/foundation-model",
            "finetune": "",
        },
        "aws_infra": {
            "instance_type": "ml.p4de.24xlarge",
            "accelerator_type": "NVIDIA_A100_80GB",
            "accelerator_count": 8,
            "min_replica_count": 1,
            "max_replica_count": 2,
            "images": {
                "serve": "public.ecr.aws/n6b1s4s0/constellaxion/serving:foundation-model-v1",
                "finetune": "",
            },
        },
    },
    "@falcon-7b-instruct": {
        "base_model": "tiiuae/falcon-7b-instruct",
        "gcp_infra": {
            "machine_type": "g2-standard-8",
            "accelerator_type": "NVIDIA_L4",
            "accelerator_count": 1,
            "replica_count": 1,
            "images": {
                "serve": "europe-west2-docker.pkg.dev/constellaxion/serving-images/foundation-model:latest",
                "finetune": "",
            },
        },
        "aws_infra": {
            "instance_type": "ml.g5.xlarge",
            "accelerator_type": "NVIDIA_L4",
            "accelerator_count": 1,
            "min_replica_count": 1,
            "max_replica_count": 2,
            "images": {
                "serve": "001939129420.dkr.ecr.eu-west-1.amazonaws.com/constellaxion/serving-images:foundation-model-v1",
                "finetune": "",
            },
        },
    },
    "@falcon-40b-instruct": {
        "base_model": "tiiuae/falcon-40b-instruct",
        "gcp_infra": {
            "machine_type": "n1-highmem-32",
            "accelerator_type": "NVIDIA_A100_40GB",
            "accelerator_count": 4,
            "replica_count": 1,
            "images": {
                "serve": "europe-west2-docker.pkg.dev/constellaxion/serving-images/foundation-model:latest",
                "finetune": "",
            },
        },
        "aws_infra": {
            "instance_type": "ml.p4d.24xlarge",
            "accelerator_type": "NVIDIA_A100_40GB",
            "accelerator_count": 4,
            "min_replica_count": 1,
            "max_replica_count": 2,
            "images": {
                "serve": "public.ecr.aws/n6b1s4s0/constellaxion/serving:foundation-model-v1",
                "finetune": "",
            },
        },
    },
    "@gemma-3-4b-it": {
        "base_model": "google/gemma-3-4b-it",
        "gcp_infra": {
            "machine_type": "g2-standard-8",
            "accelerator_type": "NVIDIA_L4",
            "accelerator_count": 1,
            "replica_count": 1,
            "images": {
                "serve": "europe-west2-docker.pkg.dev/constellaxion/serving-images/foundation-model:latest",
                "finetune": "",
            },
        },
        "aws_infra": {
            "instance_type": "ml.g5.xlarge",
            "accelerator_type": "NVIDIA_L4",
            "accelerator_count": 1,
            "min_replica_count": 1,
            "max_replica_count": 2,
            "images": {
                "serve": "public.ecr.aws/n6b1s4s0/constellaxion/serving:foundation-model-v1",
                "finetune": "",
            },
        },
    },
}



import json

import boto3
from botocore.exceptions import ClientError

ROLE_NAME = "constellaxion-admin"
POLICIES = [
    "arn:aws:iam::aws:policy/AmazonSageMakerFullAccess",
    "arn:aws:iam::aws:policy/AmazonS3FullAccess",
    "arn:aws:iam::aws:policy/AmazonEC2ContainerRegistryFullAccess",
]


def get_current_user_arn():
    """Get the ARN of the currently authenticated IAM user."""
    try:
        sts = boto3.client("sts")
        identity = sts.get_caller_identity()
        return identity["Arn"]
    except ClientError as e:
        print(
            "Error: Could not determine the current user. Are you logged in with AWS CLI?"
        )
        print(f"Details: {e}")
        raise


def add_inline_ecr_policy(iam_client):
    """Adds an inline ECR access policy to the Constellaxion IAM role."""
    ecr_inline_policy = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Action": [
                    "ecr:GetAuthorizationToken",
                    "ecr:BatchCheckLayerAvailability",
                    "ecr:GetDownloadUrlForLayer",
                    "ecr:BatchGetImage",
                ],
                "Resource": "*",
            }
        ],
    }

    try:
        iam_client.put_role_policy(
            RoleName=ROLE_NAME,
            PolicyName="ConstellaxionECRAccess",
            PolicyDocument=json.dumps(ecr_inline_policy),
        )
        print("Inline ECR access policy attached to role.")
    except ClientError as e:
        print(f"Failed to attach inline policy: {e}")
        raise


def create_iam_role():
    """Create the Constellaxion IAM role with proper trust and attach policies."""
    iam = boto3.client("iam")

    assume_role_policy_document = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Principal": {
                    "Service": ["sagemaker.amazonaws.com", "ecs-tasks.amazonaws.com"]
                },
                "Action": "sts:AssumeRole",
            }
        ],
    }

    try:
        iam.create_role(
            RoleName=ROLE_NAME,
            AssumeRolePolicyDocument=json.dumps(assume_role_policy_document),
            Description="Constellaxion Admin Role for deploying and training models",
        )
        print(f"Role '{ROLE_NAME}' created successfully.")
    except ClientError as e:
        if e.response["Error"]["Code"] == "EntityAlreadyExists":
            print(f"Role '{ROLE_NAME}' already exists. Continuing...")
        else:
            raise

    for policy_arn in POLICIES:
        try:
            iam.attach_role_policy(RoleName=ROLE_NAME, PolicyArn=policy_arn)
            print(f"Attached policy: {policy_arn}")
        except ClientError as e:
            print(f"Failed to attach policy {policy_arn}: {e}")
            raise

    # Attach inline ECR policy
    add_inline_ecr_policy(iam)


def create_aws_permissions():
    """Creates the Constellaxion IAM role and attaches necessary policies."""
    try:
        user_arn = get_current_user_arn()
        print(f"Authenticated as: {user_arn}")
        create_iam_role()
        print("IAM initialization complete.")
    except Exception as e:
        print(f"Initialization failed: {e}")



"""LoRA fine-tuning script with Unsloth optimizations for GCP deployment."""

import argparse
import inspect
import os

# Must be first non-standard import!
from unsloth import FastLanguageModel, is_bfloat16_supported

from constellaxion_utils.gcs.gcs_uploader import ModelManager
from datasets import Dataset
from google.cloud import aiplatform, storage
import pandas as pd
from transformers import TrainingArguments
from transformers.integrations import TensorBoardCallback
from trl import SFTTrainer

# Parse cli args
parser = argparse.ArgumentParser()
parser.add_argument("--epochs", type=str, required=True, help="Training epochs")
parser.add_argument("--batch-size", type=str, required=True, help="Batch size")
parser.add_argument("--train-set", type=str, required=True, help="Training set path")
parser.add_argument("--val-set", type=str, required=True, help="Validation set path")
parser.add_argument("--test-set", type=str, required=True, help="Test set path")
parser.add_argument("--dtype", type=str, required=True, help="Data type")
parser.add_argument(
    "--max-seq-length", type=str, required=True, help="Max sequence length"
)
parser.add_argument("--bucket-name", type=str, required=True, help="GCS bucket name")
parser.add_argument(
    "--model-path", type=str, required=True, help="Model artefacts output path"
)
parser.add_argument("--model-id", type=str, required=True, help="Model ID")
parser.add_argument("--base-model", type=str, required=True, help="Base model name")
parser.add_argument(
    "--experiments-dir", type=str, required=True, help="Experiments output path"
)
parser.add_argument("--location", type=str, required=True, help="Location")
parser.add_argument("--project-id", type=str, required=True, help="Project ID")
parser.add_argument(
    "--experiment-name", type=str, required=True, help="Experiment name"
)
args = parser.parse_args()

SEED = 42

LOCAL_MODEL_DIR = "./models"
CHECKPOINT_DIR = "./checkpoints"
MODEL_NAME = args.base_model
GCS_BUCKET_NAME = args.bucket_name
GCS_MODEL_PATH = args.model_path
LOCATION = args.location
PROJECT_ID = args.project_id
MODEL_ID = args.model_id
EPOCHS = args.epochs
BATCH_SIZE = args.batch_size
EXPERIMENT_NAME = args.experiment_name
EXPERIMENT_DIR = args.experiments_dir
DTYPE = args.dtype
MAX_SEQ_LENGTH = args.max_seq_length
tensorboard_path = os.environ.get("AIP_TENSORBOARD_LOG_DIR")
TRAIN_SET = f"gs://{GCS_BUCKET_NAME}/{args.train_set}"
VAL_SET = f"gs://{GCS_BUCKET_NAME}/{args.val_set}"
TEST_SET = f"gs://{GCS_BUCKET_NAME}/{args.test_set}"
OUTPUT_DIR = f"/gcs/{GCS_BUCKET_NAME}/{EXPERIMENT_DIR}"


def gcs_uri_to_fuse_path(gcs_uri: str) -> str:
    """
    Convert a gs:// URI to its FUSE-mounted /gcs/ path.

    Args:
        gcs_uri (str): A GCS path in the form gs://bucket-name/path/to/file

    Returns:
        str: The corresponding /gcs/bucket-name/path/to/file path

    Raises:
        ValueError: If the input is not a valid gs:// URI
    """
    if not gcs_uri.startswith("gs://"):
        raise ValueError(f"Invalid GCS URI: {gcs_uri}. Must start with 'gs://'")

    # Remove 'gs://' and split into bucket and path
    parts = gcs_uri[5:].split("/", 1)
    bucket = parts[0]
    path = parts[1] if len(parts) > 1 else ""

    return f"/gcs/{bucket}/{path}" if path else f"/gcs/{bucket}"


tensorboard_path = gcs_uri_to_fuse_path(tensorboard_path)

# Dataset
train_df = pd.read_csv(TRAIN_SET)
val_df = pd.read_csv(VAL_SET)
test_df = pd.read_csv(TEST_SET)

dataset = {
    "train": Dataset.from_pandas(train_df),
    "val": Dataset.from_pandas(val_df),
    "test": Dataset.from_pandas(test_df),
}

model_manager = ModelManager()
checkpoint = model_manager.prepare_checkpoint(
    GCS_BUCKET_NAME, EXPERIMENT_DIR, CHECKPOINT_DIR
)

if checkpoint:
    MODEL_PATH = checkpoint
else:
    MODEL_PATH = MODEL_NAME


# Initialize Unsloth FastLanguageModel
model, tokenizer = FastLanguageModel.from_pretrained(
    model_name=MODEL_PATH,
    max_seq_length=int(MAX_SEQ_LENGTH),
    dtype=None if not DTYPE or DTYPE == "None" else DTYPE,
    load_in_4bit=True,
)

model = FastLanguageModel.get_peft_model(
    model,
    r=32,  # Choose any number > 0 ! Suggested 8, 16, 32, 64, 128
    target_modules=[
        "q_proj",
        "k_proj",
        "v_proj",
        "o_proj",
        "gate_proj",
        "up_proj",
        "down_proj",
    ],
    lora_alpha=32,
    lora_dropout=0,  # Currently only supports dropout = 0
    bias="none",  # Currently only supports bias = "none"
    use_gradient_checkpointing=False,  # @@@ IF YOU GET OUT OF MEMORY - set to True @@@
    random_state=3407,
    use_rslora=False,  # We support rank stabilized LoRA
    loftq_config=None,  # And LoftQ
)

# Prepare model with LoRA
# model = get_peft_model(model, lora_config)

# if checkpoint:
#     model.load_state_dict(PeftModel.from_pretrained(model, MODEL_PATH).state_dict())

model.print_trainable_parameters()


EOS_TOKEN = tokenizer.eos_token


def format_prompts(example):
    """Formatter for training and validation examples"""
    output_texts = []
    for i in range(len(example["prompt"])):
        text = (
            inspect.cleandoc(
                f"""
## Prompt:
{example["prompt"][i]}
## Response:
{example["response"][i]}
"""
            )
            + EOS_TOKEN
        )
        output_texts.append(text)
    return {"text": output_texts}


# Map datasets to format_prompts
train_dataset = dataset["train"].map(
    format_prompts, batched=True, remove_columns=["prompt", "response"]
)
val_dataset = dataset["val"].map(
    format_prompts, batched=True, remove_columns=["prompt", "response"]
)

# Initialize Vertex AI with experiment tracking
aiplatform.init(
    project=PROJECT_ID,
    location=LOCATION,
    experiment=EXPERIMENT_NAME,
    experiment_description="constellaXion LoRA fine-tuning experiment",
)

# Train Model
train_args = TrainingArguments(
    per_device_train_batch_size=int(BATCH_SIZE),
    gradient_accumulation_steps=4,
    warmup_ratio=0.1,
    num_train_epochs=int(EPOCHS),
    learning_rate=2e-5,
    eval_strategy="steps",
    fp16=not is_bfloat16_supported(),
    bf16=is_bfloat16_supported(),
    logging_steps=100,
    save_strategy="steps",
    save_steps=0.2,
    optim="adamw_8bit",
    weight_decay=0.1,
    lr_scheduler_type="linear",
    seed=3407,
    output_dir=OUTPUT_DIR,
    report_to=["tensorboard"],
    logging_dir=tensorboard_path,
)

trainer = SFTTrainer(
    formatting_func=format_prompts,
    model=model,
    tokenizer=tokenizer,
    train_dataset=train_dataset,
    eval_dataset=val_dataset,
    dataset_text_field="text",
    max_seq_length=int(MAX_SEQ_LENGTH),
    dataset_num_proc=2,
    packing=True,  # Packs short sequences together to save time!
    args=train_args,
    callbacks=[TensorBoardCallback()],
)

# Train model
if checkpoint:
    trainer.train(resume_from_checkpoint=MODEL_PATH)
else:
    trainer.train()


# Upload model to GCS
def upload_directory_to_gcs(local_path, bucket_name, gcs_path):
    """Upload to GCS"""
    client = storage.Client()
    bucket = client.bucket(bucket_name)

    for root, _, files in os.walk(local_path):
        for file in files:
            local_file_path = os.path.join(root, file)
            relative_path = os.path.relpath(local_file_path, local_path)
            gcs_blob_path = os.path.join(gcs_path, relative_path)

            blob = bucket.blob(gcs_blob_path)
            blob.upload_from_filename(local_file_path)
            print(
                f"Uploaded {local_file_path} to " f"gs://{bucket_name}/{gcs_blob_path}"
            )


def save_model_tokenizer_locally(m, t, save_dir):
    """Save model and tokenizer locally"""
    os.makedirs(save_dir, exist_ok=True)
    m.save_pretrained(save_dir)
    t.save_pretrained(save_dir)
    print(f"Model and tokenizer saved locally to {save_dir}")


def save_and_upload_model(m, t):
    """Save and upload model"""
    # Save locally
    save_model_tokenizer_locally(m, t, LOCAL_MODEL_DIR)

    # Upload to GCS
    upload_directory_to_gcs(LOCAL_MODEL_DIR, GCS_BUCKET_NAME, GCS_MODEL_PATH)


save_and_upload_model(trainer.model, tokenizer)



import json

from google.cloud import aiplatform, storage
import pkg_resources

from constellaxion.models.model_map import model_map


def create_vertex_dataset(
    model_id: str,
    bucket_name: str,
    train_set: str,
    val_set: str,
    test_set: str,
    location: str,
) -> None:
    """
    Checks if a default metadata store exists in the project at the specified location.
    Creates one if it doesn't exist.

    Args:
        project_id (str): GCP project ID
        location (str): GCP region (e.g., 'us-central1')
    """
    try:
        # Create GCS paths
        gcs_paths = [
            f"gs://{bucket_name}/{train_set}",
            f"gs://{bucket_name}/{val_set}",
            f"gs://{bucket_name}/{test_set}",
        ]

        # Create dataset without using paths as label values
        aiplatform.TabularDataset.create(
            display_name=f"{model_id}-dataset",
            location=location,
            gcs_source=gcs_paths,
            labels={"model_id": model_id},
        )

    except Exception as e:
        print(f"Warning: Error while checking/creating metadata store: {str(e)}")
        print("Continuing with training...")


def upload_data_to_gcp(config: dict):
    """
    Upload dataset to GCP, ensuring the bucket exists.

    Args:
        config (dict): Configuration dictionary with bucket and dataset details.
    """
    client = storage.Client()
    bucket_name = config["deploy"]["bucket_name"]

    # Check if bucket exists
    bucket = client.bucket(bucket_name)
    if not bucket.exists():
        print(f"Bucket '{bucket_name}' does not exist. Creating it...")
        bucket = client.create_bucket(bucket_name)
        print(f"Bucket '{bucket_name}' created successfully.")

    # Upload training dataset
    train_blob = bucket.blob(config["dataset"]["train"]["cloud"])
    train_blob.upload_from_filename(config["dataset"]["train"]["local"])
    print(f"Uploaded training dataset to {train_blob.name}")

    # Upload validation dataset
    val_blob = bucket.blob(config["dataset"]["val"]["cloud"])
    val_blob.upload_from_filename(config["dataset"]["val"]["local"])
    print(f"Uploaded validation dataset to {val_blob.name}")

    # Upload test dataset
    test_blob = bucket.blob(config["dataset"]["test"]["cloud"])
    test_blob.upload_from_filename(config["dataset"]["test"]["local"])
    print(f"Uploaded test dataset to {test_blob.name}")


def create_training_job(
    project: str,
    location: str,
    staging_bucket: str,
    display_name: str,
    script_path: str,
    container_uri: str,
    service_account: str,
    # requirements: str,
    machine_type: str,
    accelerator_type: str,
    accelerator_count: int,
    replica_count: int,
    experiment_name: str,
    args: list[str],
) -> None:
    """Creates and runs a Vertex AI custom training job with TensorBoard integration."""
    aiplatform.init(project=project, location=location, staging_bucket=staging_bucket)

    # Try to get existing experiment, create if it doesn't exist
    try:
        experiment = aiplatform.Experiment(experiment_name)
    except Exception:
        experiment = aiplatform.Experiment.create(experiment_name)

    # Get TensorBoard instance
    tensorboard = experiment.get_backing_tensorboard_resource()
    if tensorboard is None:
        # Create a new TensorBoard instance if one doesn't exist
        tensorboard = aiplatform.Tensorboard.create(
            display_name=f"{experiment_name}-tensorboard",
            project=project,
            location=location,
        )
        # Associate the tensorboard with the experiment
        experiment.assign_backing_tensorboard(tensorboard)

    tensorboard_resource_name = tensorboard.gca_resource.name
    print(f"Original Tensorboard resource name: {tensorboard_resource_name}")
    # Extract experiment ID from the resource name
    experiment_id = experiment.resource_name.split("/")[-1]

    # Parse the tensorboard resource name to get project ID and tensorboard ID
    parts = tensorboard_resource_name.split("/")
    project_number = parts[1]  # Gets the numeric project ID
    tensorboard_id = parts[-1]

    # Generate Tensorboard URL in the correct format
    tensorboard_url = (
        f"https://{location}.tensorboard.googleusercontent.com/experiment/"
        f"projects+{project_number}+locations+{location}+tensorboards+{tensorboard_id}"
        f"+experiments+{experiment_id}/#scalars"
    )
    print(f"Tensorboard URL: {tensorboard_url}")

    # Read existing job.json
    try:
        with open("job.json", "r", encoding="utf-8") as f:
            job_config = json.load(f)
    except FileNotFoundError:
        job_config = {}

    # Ensure 'training' key exists
    if "training" not in job_config:
        job_config["training"] = {}

    # Add tensorboard URL to training section
    job_config["training"]["tensorboard_url"] = tensorboard_url

    # Write updated config back to job.json
    with open("job.json", "w", encoding="utf-8") as f:
        json.dump(job_config, f, indent=4)

    job = aiplatform.CustomJob.from_local_script(
        display_name=display_name,
        script_path=script_path,
        container_uri=container_uri,
        # requirements=requirements,
        machine_type=machine_type,
        accelerator_type=accelerator_type,
        accelerator_count=accelerator_count,
        replica_count=replica_count,
        args=args,
    )
    print(f"Tensorboard resource name: {tensorboard_resource_name}")
    job.run(service_account=service_account, tensorboard=tensorboard_resource_name)


def run_training_job(config):
    """
    Run a training job for a given model.

    Args:
        config (dict): Configuration dictionary with model and dataset details.
    """
    base_model_alias = config["model"]["base_model"]
    bucket_name = config["deploy"]["bucket_name"]
    model_id = config["model"]["model_id"]
    train_set = config["dataset"]["train"]["cloud"]
    val_set = config["dataset"]["val"]["cloud"]
    test_set = config["dataset"]["test"]["cloud"]
    script_path = pkg_resources.resource_filename(
        "constellaxion.models.scripts.gcp", "lora.py"
    )
    infra_config = model_map[base_model_alias]["gcp_infra"]
    dtype = model_map[base_model_alias]["dtype"]
    max_seq_length = model_map[base_model_alias]["max_seq_length"]
    base_model = model_map[base_model_alias]["base_model"]
    # Upload data to GCP
    upload_data_to_gcp(config)
    project_id = config["deploy"]["project_id"]
    location = config["deploy"]["region"]
    experiment_name = (
        f"{config['model']['model_id']}-lora-{config['training']['epochs']}e"
    )
    # Add this before initializing the experiment
    create_vertex_dataset(
        experiment_name, bucket_name, train_set, val_set, test_set, location
    )
    create_training_job(
        project=project_id,
        location=location,
        staging_bucket=f"gs://{bucket_name}/{config['deploy']['staging_dir']}",
        display_name=config["model"]["model_id"],
        script_path=script_path,
        # requirements=finetune_packages,
        container_uri=infra_config["images"]["finetune"],
        service_account=config["deploy"]["service_account"],
        machine_type=infra_config["machine_type"],
        accelerator_type=infra_config["accelerator_type"],
        accelerator_count=infra_config["accelerator_count"],
        replica_count=infra_config["replica_count"],
        experiment_name=experiment_name,
        args=[
            f"--epochs={config['training']['epochs']}",
            f"--batch-size={config['training']['batch_size']}",
            f"--train-set={config['dataset']['train']['cloud']}",
            f"--val-set={config['dataset']['val']['cloud']}",
            f"--test-set={config['dataset']['test']['cloud']}",
            f"--dtype={dtype}",
            f"--max-seq-length={max_seq_length}",
            f"--base-model={base_model}",
            f"--bucket-name={bucket_name}",
            f"--model-path={config['deploy']['model_path']}",
            f"--experiments-dir={config['deploy']['experiments_dir']}",
            f"--location={location}",
            f"--project-id={project_id}",
            f"--model-id={model_id}",
            f"--experiment-name={experiment_name}",
        ],
    )



"""Initialize a new model configuration and setup required resources."""

import os

import click
from halo import Halo
from rich.console import Console
from rich.panel import Panel
from rich.text import Text
from tabulate import tabulate
import yaml

from constellaxion.handlers.cloud_job import AWSDeployJob, GCPDeployJob
from constellaxion.handlers.dataset import Dataset
from constellaxion.handlers.model import Model
from constellaxion.handlers.training import Training
from constellaxion.services.aws.iam import create_aws_permissions
from constellaxion.services.gcp.iam import create_service_account

console = Console()

CONSTELLAXION_LOGO = """\
░█▀▀░█▀█░█▀█░█▀▀░▀█▀░█▀▀░█░░░█░░░█▀█░█░█░▀█▀░█▀█░█▀█
░█░░░█░█░█░█░▀▀█░░█░░█▀▀░█░░░█░░░█▀█░▄▀▄░░█░░█░█░█░█
░▀▀▀░▀▀▀░▀░▀░▀▀▀░░▀░░▀▀▀░▀▀▀░▀▀▀░▀░▀░▀░▀░▀▀▀░▀▀▀░▀░▀
"""


CXN_LOGO = """\
 ██████╗██╗  ██╗███╗   ██╗
██╔════╝╚██╗██╔╝████╗  ██║
██║      ╚███╔╝ ██╔██╗ ██║
██║      ██╔██╗ ██║╚██╗██║
╚██████╗██╔╝ ██╗██║ ╚████║
 ╚═════╝╚═╝  ╚═╝╚═╝  ╚═══╝
"""


def init_model(model_config):
    """Initialize the model

    Args:
        model_config (dict): Model config details

    Returns:
        Model: Initialized model instance

    Raises:
        AttributeError: If required attributes (id or base) are missing
    """
    model_id = model_config.get("id")
    base = model_config.get("base")
    if not model_id:
        raise AttributeError("Missing value, model.id in model.yaml file")
    if not base:
        raise AttributeError("Missing value, model.base in model.yaml file")
    return Model(model_id, base)


def init_dataset(dataset_config, model_config):
    """Initialize the dataset

    Args:
        dataset_config (dict): Dataset config details
    """
    model_id = model_config.get("id")
    train = dataset_config.get("train")
    val = dataset_config.get("val")
    test = dataset_config.get("test")
    if not train:
        raise AttributeError("Missing value, dataset.train in model.yaml file")
    if not val:
        raise AttributeError("Missing value, dataset.val in model.yaml file")
    if not test:
        raise AttributeError("Missing value, dataset.test in model.yaml file")
    return Dataset(train, val, test, model_id)


def init_training(training_config):
    """Initialize the dataset

    Args:
        dataset_config (dict): Dataset config details
    """
    epochs = training_config.get("epochs")
    batch_size = training_config.get("batch_size")
    if not epochs:
        raise AttributeError("Missing value, training.epochs in model.yaml file")
    if not batch_size:
        raise AttributeError("Missing value, training.batch_size in model.yaml file")
    return Training(epochs, batch_size)


def init_job(job_config, model: Model, dataset: Dataset, training: Training):
    """Initialize the deployment job definition

    Args:
        job_config (list): List of dicts containing deployment job config details
    """
    platform = None
    if "aws" in job_config:
        platform = "aws"
    elif "gcp" in job_config:
        platform = "gcp"
    else:
        click.echo(
            "Error: Missing value, job.gcp or job.aws in model.yaml file", err=True
        )
    # Initialize GCP resources
    if platform == "gcp":
        gcp = job_config.get("gcp")
        project_id = gcp.get("project_id")
        region = gcp.get("region")
        if not project_id:
            click.echo(
                "Error: Missing value, job.gcp.project_id in model.yaml file", err=True
            )
        if not region:
            click.echo(
                "Error: Missing value, job.gcp.region in model.yaml file", err=True
            )

        click.echo(f"Initializing resources for project: {project_id}")
        try:
            service_account_email = create_service_account(project_id)
            if service_account_email:
                click.echo("The required GCP Service Account is ready to use 🦾")
        except (ValueError, RuntimeError) as e:
            click.echo(f"Error: {str(e)}", err=True)
        job = GCPDeployJob()
        # Create job config
        job.create_config(
            model, project_id, region, service_account_email, dataset, training
        )

    # Initialize AWS resources
    elif platform == "aws":
        aws = job_config.get("aws")
        region = aws.get("region")
        if not region:
            raise ValueError("Missing value, job.aws.region in model.yaml file")
        create_aws_permissions()
        job = AWSDeployJob()
        # Create job config
        job.create_config(model, region, dataset, training)


def show_after_init_command_table():
    """Show the command table"""
    table = [
        ["Command", "Description"],
        ["constellaXion model view", "View the current model configuration"],
        ["constellaXion model train", "Run finetuning job"],
        ["constellaXion model deploy", "Deploy a foundation model"],
    ]
    click.echo(tabulate(table, headers="firstrow", tablefmt="grid"))


@click.command(help="Initialize a new model")
def init():
    """
    Initialize a new model
    """
    # Print the logo
    console.print(
        Panel(Text(CONSTELLAXION_LOGO, justify="center"), style="#47589B", expand=True)
    )

    # Start loading animation
    spinner = Halo(spinner="dots")
    spinner.start()

    # Load the model config
    model_config = os.path.join(os.getcwd(), "model.yaml")
    if not os.path.exists(model_config):
        click.echo("Error: model.yaml file not found in current directory.", err=True)
        raise click.Abort()

    click.echo("Preparing new model config 📡")
    try:
        with open(model_config, "r", encoding="utf-8") as file:
            config = yaml.safe_load(file)
            training = None
            dataset = None
            # Get configs
            model_config = config.get("model")
            training_config = config.get("training")
            # If training config is present, initialize training
            if training_config:
                training = init_training(training_config)
                dataset_config = config.get("dataset")
                # Ensure dataset config is present if training config is present
                if not dataset_config:
                    click.echo(
                        "Error: Missing value, dataset in model.yaml file", err=True
                    )
                    raise click.Abort()
                dataset = init_dataset(dataset_config, model_config)
            deploy_config = config.get("deploy")
            if not deploy_config:
                click.echo("Error: Missing value, deploy in model.yaml file", err=True)
                raise click.Abort()
            # Init configs
            model = init_model(model_config)
            init_job(deploy_config, model, dataset, training)

            spinner.succeed("Initialization complete!")
            click.echo(click.style("Job Config created", fg="green"))
            show_after_init_command_table()
    # Parse values and excecute commands
    except yaml.YAMLError as e:
        click.echo(f"Error parsing model.yaml: {str(e)}", err=True)
        raise click.Abort()
    except (OSError, ValueError, KeyError) as e:
        click.echo(f"Error: {str(e)}", err=True)
        raise click.Abort()



class Training:
    """Training class for handling training parameters."""

    def __init__(self, epochs: str, batch_size):
        if not epochs or not batch_size:
            raise ValueError("Epochs and batch size must be provided")
        self.epochs = epochs
        self.batch_size = batch_size

    def to_dict(self):
        """Convert the training to a dictionary."""
        return {
            "epochs": self.epochs,
            "batch_size": self.batch_size,
        }



"""Model class for handling model configuration."""


class Model:
    """Represents a model with its ID and base model configuration."""

    def __init__(self, model_id: str, base_model: str):
        if not model_id or not base_model:
            raise ValueError("model must have an id and base in model.yaml file")
        self.id = model_id
        self.base_model = base_model



import sys

from setuptools import find_packages, setup

if sys.platform == "win32":
    print("NOTE: For Windows, ensure Microsoft C++ Build Tools are installed.")

setup(
    name="constellaxion",
    version="0.4.1",
    packages=find_packages(),
    install_requires=open("requirements.txt", encoding="utf-8").read().splitlines(),
    entry_points={
        "console_scripts": [
            "constellaxion=constellaxion.main:cli",
        ],
    },
    author="Constellaxion Technologies, Inc.",
    author_email="dev@constellaxion.ai",
    description="The constellaXion CLI for managing your laboratory database",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    keywords=["constellaxion", "ai", "ml ops"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
    package_data={"constellaxion": ["services/gcp/*.py", "models/scripts/**/*.py"]},
)



"""
This module provides a class for uploading files to a GCS bucket.
"""

import os
import threading
import time

import gcsfs
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer


class GCSUploaderHandler(FileSystemEventHandler):
    """
    A class that handles file system events and uploads them to a GCS bucket.
    """

    def __init__(self, local_dir, gcs_dir):
        self.local_dir = local_dir
        self.gcs_dir = gcs_dir
        self.fs = gcsfs.GCSFileSystem()

    def on_modified(self, event):
        if not event.is_directory:
            self.upload_file(event.src_path)

    def on_created(self, event):
        if not event.is_directory:
            self.upload_file(event.src_path)

    def upload_file(self, file_path):
        """Upload a file to a GCS bucket."""
        relative_path = os.path.relpath(file_path, self.local_dir)
        gcs_path = os.path.join(self.gcs_dir, relative_path)

        try:
            with open(file_path, "rb") as f:
                with self.fs.open(gcs_path, "wb") as gcs_file:
                    gcs_file.write(f.read())
            print(f"✅ Uploaded: {relative_path} to {gcs_path}")
        except Exception as e:
            print(f"❌ Failed to upload {relative_path}: {e}")

    def upload_directory(self, directory_path):
        """Upload a directory to a GCS bucket."""
        for root, dirs, files in os.walk(directory_path):
            for file in files:
                self.upload_file(os.path.join(root, file))
            for directory in dirs:
                self.upload_directory(os.path.join(root, directory))


def start_gcs_sync(local_dir, gcs_dir):
    """Sync local directory to GCS bucket"""
    event_handler = GCSUploaderHandler(local_dir, gcs_dir)
    observer = Observer()
    observer.schedule(event_handler, path=local_dir, recursive=True)
    observer.start()
    print(f"🚀 Started syncing from {local_dir} to {gcs_dir}")

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()


def start_gcs_sync_thread(local_dir, gcs_dir):
    """Start the GCS sync in a separate thread"""
    os.makedirs(local_dir, exist_ok=True)
    sync_thread = threading.Thread(
        target=start_gcs_sync, args=(local_dir, gcs_dir), daemon=True
    )
    sync_thread.start()



import click

from constellaxion.commands.init import init
from constellaxion.commands.login import login
from constellaxion.commands.model import model


@click.group()
def cli():
    """Constellaxion CLI: Infrastructure deployment for LLMs"""
    pass


cli.add_command(login)
cli.add_command(init)
cli.add_command(model)

if __name__ == "__main__":
    cli()



from google.cloud import aiplatform

from constellaxion.models.model_map import model_map


def create_model_from_custom_container(model_name: str, image_uri: str, env_vars: dict):
    """Creates a Vertex AI model from a custom container image."""
    print("Creating model from custom container...")

    # Define the container and model
    model = aiplatform.Model.upload(
        display_name=model_name,
        serving_container_image_uri=image_uri,
        serving_container_ports=[8080],  # Port your container exposes
        # Optional, adjust if your container provides health checks
        serving_container_health_route="/health",
        # Optional, adjust if your container provides a prediction endpoint
        serving_container_predict_route="/predict",
        serving_container_environment_variables=env_vars,
    )
    print(f"Model created successfully: {model.resource_name}")
    return model


# Deploy the model to an endpoint


def deploy_model_to_endpoint(
    model,
    model_id: str,
    machine_type: str,
    accelerator_type: str,
    accelerator_count: int,
    replica_count: int,
    service_account: str,
):
    """Deploys a model to a Vertex AI endpoint with specified compute resources."""
    # Check if the endpoint exists, create it if not
    endpoints = aiplatform.Endpoint.list(filter=f'display_name="{model_id}"')
    if endpoints:
        endpoint = endpoints[0]  # Use the first matching endpoint
        print(f"Using existing endpoint: {endpoint.display_name}")
    else:
        # Create a new endpoint
        endpoint = aiplatform.Endpoint.create(display_name=model_id)
        print(f"Created new endpoint: {endpoint.display_name}")

    # Deploy the model to the endpoint
    model.deploy(
        endpoint=endpoint,
        deployed_model_display_name=model_id,
        traffic_split={"0": 100},  # Route all traffic to this model
        accelerator_type=accelerator_type,
        accelerator_count=accelerator_count,
        machine_type=machine_type,
        min_replica_count=replica_count,
        max_replica_count=replica_count,
        service_account=service_account,
    )

    print(f"Model deployed to endpoint: {endpoint.resource_name}")
    return endpoint.resource_name


def run_gcp_deploy_job(config):
    """Runs the GCP deployment job by creating and deploying a model to Vertex AI."""
    project_id = config["deploy"]["project_id"]
    base_model_alias = config["model"]["base_model"]
    region = config["deploy"]["region"]
    model_id = config["model"]["model_id"]
    service_account = config["deploy"]["service_account"]
    infra_config = model_map[base_model_alias]["gcp_infra"]
    base_model = model_map[base_model_alias]["base_model"]
    image_uri = infra_config["images"]["serve"]
    machine_type = infra_config["machine_type"]
    accelerator_type = infra_config["accelerator_type"]
    accelerator_count = infra_config["accelerator_count"]
    replica_count = infra_config["replica_count"]
    dtype = infra_config["dtype"] if "dtype" in infra_config else None
    env_vars = {"MODEL_NAME": base_model, "DTYPE": dtype}
    # Initialize the Vertex AI SDK
    aiplatform.init(project=project_id, location=region)
    # # Download the model
    model = create_model_from_custom_container(base_model, image_uri, env_vars)
    # Deploy model to endpoint
    endpoint_path = deploy_model_to_endpoint(
        model,
        model_id,
        machine_type,
        accelerator_type,
        accelerator_count,
        replica_count,
        service_account,
    )
    return endpoint_path



"""Module for sending prompts to SageMaker endpoints and receiving responses."""

import json

import boto3


def send_aws_prompt(prompt: str, endpoint_name: str, region: str):
    """Sends a request to a SageMaker endpoint and returns the response."""
    try:
        sagemaker_runtime = boto3.client("sagemaker-runtime", region_name=region)
        body = bytes(f'{{"inputs": "{prompt}"}}', "utf-8")
        response = sagemaker_runtime.invoke_endpoint(
            EndpointName=endpoint_name, Body=body, ContentType="application/json"
        )
        payload = response["Body"].read().decode("utf-8")
        text = json.loads(payload)["generated_text"]
    except (boto3.exceptions.Boto3Error, json.JSONDecodeError, KeyError) as e:
        print(f"Error sending prompt to SageMaker endpoint: {e}")
        return None

    return text
